
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _GNU_SOURCE
#include <fcntl.h>
#include <sys/file.h>
#include <sys/types.h>
#include <signal.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <unistd.h>
#include <sys/time.h>
#include <time.h>
#include <getopt.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include <errno.h>

#ifdef HAVE_NCURSES_H
#include <term.h>
#include <ncurses.h>
#endif
#ifdef HAVE_NCURSES_NCURSES_H
#include <ncurses/term.h>
#include <ncurses/ncurses.h>
#endif

/* #include "ngc.h" */
#include <initng.h>
#include <initng_global.h>
#include <initng_control_command.h>
#include <initng_active_state.h>
#include <initng_toolbox.h>
#include <initng-paths.h>

#include "initng_ngc2.h"

#include "ansi_colors.h"

/* maximum size of string received, warning ngc -p can contain a lot */
#define MAX_RESCIVE_SIZE 100000

#define TRUE 1
#define FALSE 0
int debug = FALSE;

int header_printed = FALSE;
int try_count = 0;
int quiet = FALSE;
int nrows = 0;
int ncolumns = 0;

#define print_out(...); { if(quiet==FALSE) printf(__VA_ARGS__ ); }

#define C_ERROR "\n  "C_FG_RED" [ERROR] -->> "C_OFF

#define TMP_QUIET if(quiet==FALSE) quiet=3;
#define TMP_UNQUIET if(quiet==3) quiet=FALSE;

/* Global communication socket */
static int sock = -1;

/* functions */
void print_usage(void);
int open_socket(void);
int close_socket(void);
static result_desc *send_command(const char c, const char *l, const char *o);
static int send_and_handle(const char c, const char *l, const char *o);
void command_fetch_string(void);
int command_fetch_int(void);
int main(int argc, char *argv[]);
int ngc2_recvall(int sock, void *buf, int len);

int ngc_hlp(void);
int ngc_active_db(void);
int ngc_option_db(void);

void bad_syntax(void);
char *Argv;


#define FILL_SPACES(string, no) { \
    int t=0; \
    while(t<no) \
    { \
	if((string)[t]=='\0') \
	{ \
	    (string)[t]=' '; \
	    (string)[t+1]='\0'; \
	} \
	t++; \
    } \
}

/* print usage on screen */
void print_usage(void)
{
    send_and_handle('h', NULL, NULL);
}

/* open socket to communicate with */
int open_socket()
{
    int len;
    struct sockaddr_un sockname;

    /* Create the socket. */
    sock = socket(PF_UNIX, SOCK_STREAM, 0);
    if (sock < 0)
    {
        printf(C_ERROR "Failed to init socket" C_OFF "\n");
        return (FALSE);
    }

    /* Bind a name to the socket. */
    sockname.sun_family = AF_UNIX;
    if (debug)
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_TEST);
    else
        len = sprintf(sockname.sun_path, SOCKET_FILENAME_REAL);

    len += sizeof(sockname.sun_family);

    if (connect(sock, (struct sockaddr *) &sockname, len) < 0)
    {
        close(sock);
        printf(C_ERROR "Error connecting to initng socket\n");
        printf("This probably means that you're not running initng." C_OFF
               "\n");
        exit(1);
        return (FALSE);
    }
    return (TRUE);
}


/* close pipes */
int close_socket(void)
{
    if (sock != -1)
    {
        close(sock);
        sock = -1;
    }
    return (sock);
}



/* send a command */
static result_desc *send_command(const char c, const char *l, const char *o)
{

    read_header header;
    result_desc *result = NULL;

    /* clear structure just in case */
    memset(&header, 0, sizeof(header));
    header.p_ver = PROTOCOL_VERSION;

    /* fill the header with data */
    header.c = c;
    if (l)
        strncpy(header.l, l, 100);
    else
        header.l[0] = '\0';
	
	
    if (o)
	header.body_len =  strlen(o) * sizeof(char);


    /*print_out("Sending: %c, %s, %s\n", c, l ,o); */
    if ((sock < 0) && (open_socket() == FALSE))
    {
        printf(C_ERROR "Socket is not open, or failed to open!" C_OFF "\n");
        exit(1);
    }

    if (write(sock, &header, sizeof(read_header)) <
        (signed) sizeof(read_header))
    {
        printf(C_ERROR "Could not send header!" C_OFF "\n");
        exit(1);
    }
    
    if (write(sock, o, header.body_len) <
        (signed) header.body_len)
    {
        printf(C_ERROR "Could not send body!" C_OFF "\n");
        exit(1);
    }
    

    /* no idea to stress initng */
    usleep(50000);

    /* if we are reloading initng, we wont have an answer from it, so don't fetch any */
    if (header.c == 'c')
    {
        print_out(C_FG_YELLOW
                  "Can't give you feedback if reloading of initng succeeds or not."
                  C_OFF "\n");
	print_out("If ngc -c fails, and initng stops responing, try # init U\n");
        exit(0);
    }

    /* FETCH THE RESULT */

    /* calloc the result space */
    result = calloc(1, sizeof(result_desc));

    /* read header data */
    if (TEMP_FAILURE_RETRY(read(sock, result, sizeof(result_desc))) !=
        sizeof(*result))
    {
        printf(C_ERROR "failed to fetch the result" C_OFF "\n");
        exit(1);
    }

    /* check that protocol matches */
    if (result->p_ver != PROTOCOL_VERSION)
    {
        printf(C_ERROR "PROTOCOL_VERSION mismatch :\n");
        printf(" initng_protocol_version: %i, client_protocol_version: %i !\n", result->p_ver, PROTOCOL_VERSION);
        printf(" initng_version: %s, ngc version: %s" C_OFF "\n",
               result->version, VERSION);
        exit(1);
    }

    if (header_printed == FALSE)
    {
        /*print_out("Got an response from initng, version: %s\n", result->version); */

        if (strcmp(result->version, VERSION) != 0)
        {
            printf(C_FG_YELLOW "Version of initng and version of ngc differs!"
                   C_OFF "\n");
            printf("this might be a bad thing, that can in worst case cause problems!\n");
        }

        /*print_out("got an read: command %c, status %i, type %i\n", result->c,
           result->s, result->t); */
        header_printed = TRUE;
    }

    switch (result->s)
    {
        case S_FALSE:
            print_out("\nThe result of '-%c' is " C_FG_LIGHT_RED "[FALSE]"
                      C_OFF " = %i!\n\n", result->c, result->s);
            free(result);
            return (NULL);

        case S_REQUIRES_OPT:
            {
                if (result->c != '\0')
                {
                    printf(C_ERROR "The command '-%c' requires an option.\n"
                           C_OFF "\n", result->c);
                }
                else
                {
                    printf(C_ERROR "This command needs an option." C_OFF
                           "\n");
                }
                free(result);
                return (NULL);
            }
        case S_NOT_REQUIRES_OPT:
            printf(C_ERROR "The command '-%c' cant have an option.\n" C_OFF
                   "\n", result->c);
            free(result);
            return (NULL);

        case S_INVALID_TYPE:
            printf(C_ERROR
                   "The data returning of this command ( '-%c' ) is an unknown type.\n"
                   C_OFF "\n", result->c);
            free(result);
            return (NULL);

        case S_COMMAND_NOT_FOUND:
            {
                printf(C_ERROR "Command not found.\n" C_OFF "\n");
                free(result);
                print_usage();
                return (NULL);
            }
        case S_TRUE:
            return (result);
    }
    return (NULL);
}

/* return the fetched string */
void command_fetch_string(void)
{
    char tmp_read[MAX_RESCIVE_SIZE + 1];
    ssize_t len;

    while ((len =
            TEMP_FAILURE_RETRY(read(sock, &tmp_read, MAX_RESCIVE_SIZE))) > 0)
    {
        tmp_read[len] = '\0';
        print_out("%s\n", tmp_read);
    }

    /* TODO: error handling??? */

    close_socket();
}

/* return the fetched return number */
int command_fetch_int(void)
{
    char got_from_read[20];

    got_from_read[read(sock, &got_from_read, 19)] = '\0';
    close_socket();
    return (atoi(got_from_read));
}

static int ngc_start_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;

    print_out(C_FG_NEON_GREEN "Starting service: " C_OFF "\n");

    while (counter < 30)
    {
        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            printf(C_ERROR "Failed to fetch result" C_OFF "\n");
            close_socket();
            return FALSE;
        }

        close_socket();
        /*print_out("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */
        if (row.i == IS_UP || row.i == IS_WAITING || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*print_out("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);

        /* make sure we got one */
        if (!result || result->c != 's')
        {
            printf(C_ERROR "Invalid result!" C_OFF "\n");
            close_socket();
            return (FALSE);
        }

        counter++;
    }
    close_socket();

    if (row.i == IS_UP)
    {
        if (strcmp(row.s, "ALREADY_RUNNING") == 0)
        {
            print_out("\tService \"" C_FG_NEON_GREEN "%s" C_OFF "\" already "
                      C_FG_NEON_GREEN "started" C_OFF "!\n", row.n);
        }
        else
        {
            print_out("\tService \"" C_FG_NEON_GREEN "%s" C_OFF "\" started "
                      C_FG_NEON_GREEN "successfully" C_OFF "! (%s)\n", row.n,
                      row.s);
        }
        return (TRUE);
    }

    if (row.i == IS_WAITING)
    {
        print_out("\tService \"" C_FG_CYAN "%s" C_OFF "\" is " C_FG_CYAN
                  "starting" C_OFF "! (%s)\n", row.n, row.s);
        return (TRUE);
    }

    if (counter == 10)
    {
        print_out(C_FG_YELLOW
                  "Timeout on start service, might be started soon anyway."
                  C_OFF "\n");
        return (FALSE);
    }

    if (row.i == IS_FAILED && try_count <= 2)
    {
        try_count++;
        print_out("\n\n");
        print_out("Warning, service \"" C_FG_RED "%s" C_OFF
                  "\" is in status: \"" C_FG_RED "%s" C_OFF "\"\n\n", row.n,
                  row.s);

        if (strcmp(row.s, "NOT_FOUND") == 0)
        {
            printf(C_ERROR "This means that the service could not start,\n");
            printf("Please check initng console, syslogs or ngc -L\n");
            printf("For more info.\n" C_OFF "\n");

            return (FALSE);
        }

        /* If found but failed */
        printf(C_ERROR "Service \"%s\" was failed, but found.\n", row.n);
        printf("Will not make an attempt to reset service state and try again"
               C_OFF "\n");

        TMP_QUIET;                          /* don't print zap result */
        /* zap the service */
        if (!send_and_handle('z', NULL, row.n))
            return (FALSE);

        TMP_UNQUIET;


        sleep(1);
        print_out("\n\nTrying to start \"" C_FG_YELLOW "%s" C_OFF
                  "\" again ...\n\n", row.n);

        /* try start again */
        if (send_and_handle('u', NULL, row.n))
            return (TRUE);
    }

    printf(C_ERROR "Failed to start service, \"%s\" (%s)" C_OFF "\n", row.n,
           row.s);
    return (FALSE);
}


static int ngc_stop_result(result_desc * res)
{
    active_row row;
    result_desc *result = NULL;
    int counter = 0;
    int once_found = 0;

    print_out(C_FG_LIGHT_BLUE "\nStopping service: " C_OFF "\n");

    if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
        sizeof(row))
    {
        printf(C_ERROR "Failed to read response" C_OFF "\n");
        close_socket();
        return FALSE;
    }

    close_socket();


    while (counter < 30)
    {
        /*print_out("row.s = %s, row.n = %s, row.t = %i, row.i = %i\n", row.s, row.n, (int) row.t, row.i); */

        if (row.i == IS_DOWN || row.i == IS_FAILED)
            break;

        if (strlen(row.n) < 2)
            break;

        /* sleep awhile */
        sleep(1);

        /* And see if status has changed */
        /*print_out("Fetch result on %s\n", row.n); */
        result = send_command('s', NULL, row.n);


        /* make sure we got one */
        if (!result || result->c != 's')
        {
            printf(C_ERROR "Invalid result!" C_OFF "\n");
            close_socket();
            return (FALSE);
        }

        if (TEMP_FAILURE_RETRY(read(sock, &row, sizeof(active_row))) !=
            sizeof(row))
        {
            printf(C_ERROR "\n\tfailed to read response" C_OFF "\n");
            close_socket();
            return FALSE;
        }

        /* not found can be good */
        if (once_found == 1 && strcmp(row.s, "NOT_FOUND") == 0)
            break;

        /* If its NOT_FOUND in future, then its a Godot thing */
        once_found = 1;

        /* break if failed */
        if (row.i == IS_FAILED)
            break;

        close_socket();

        counter++;
    }
    close_socket();

    /* not found can be Good */
    if (once_found == 1 && strcmp(row.s, "NOT_FOUND") == 0)
    {
        print_out("\tService \"" C_FG_LIGHT_BLUE "%s" C_OFF "\" stopped "
                  C_FG_LIGHT_BLUE "SUCCESSFULLY" C_OFF "!\n", row.n);
        return (TRUE);
    }

    /* or bad */
    if (strcmp(row.s, "NOT_FOUND") == 0)
    {
        print_out("\n\n\tService \"" C_FG_RED "%s" C_OFF "\" does " C_FG_RED
                  "NOT_EXIST" C_OFF ".\n", row.n);
        return (TRUE);
    }


    /* Down is really good to */
    if (row.i == IS_DOWN)
    {
        print_out("\tService \"" C_FG_LIGHT_BLUE "%s" C_OFF "\" is "
                  C_FG_LIGHT_BLUE "STOPPED" C_OFF "!\n", row.n);
        return (TRUE);
    }


    printf(C_ERROR "Service \"%s\" has status failed to stop.\n", row.n);
    printf("Instead you should kill this service manually,\n");
    printf("And after that do ngc -z %s, to release\n", row.s);
    printf("its status from initng.\n" C_OFF "\n");
    return (FALSE);
}


static int handle_result(result_desc * result)
{
    /* special handling commands */
    switch (result->c)
    {
        case 'h':                          /* standard option list */
        case 'H':                          /* advanced option list */
            return (ngc_hlp());
        case 's':
            return (ngc_active_db());
        case 'O':
            return (ngc_option_db());
        case 'L':
            return (ngc_active_db());
        case 'u':
            return (ngc_start_result(result));
        case 'd':
            return (ngc_stop_result(result));
    }

    /* handle by type */
    switch (result->t)
    {

        case TRUE_OR_FALSE_COMMAND:
            {
                int r = command_fetch_int();

                if (r == TRUE)
                {
                    print_out(" command '-%c' returned " C_FG_NEON_GREEN
                              "SUCCESS" C_OFF "! \n", result->c);
                }
                else
                {
                    print_out(" command '-%c' returned " C_FG_RED "FAILURE"
                              C_OFF "! \n", result->c);
                }

                return (r);
            }
        case INT_COMMAND:
            print_out("INT_COMMAND_RESULT: " C_FG_WHITE "%i" C_OFF "\n",
                      command_fetch_int());
            return (TRUE);
        case STRING_COMMAND:
            command_fetch_string();
            return (TRUE);
        case VOID_COMMAND:
            return (TRUE);
        case COMMAND_FAIL:
        default:
            close_socket();
            printf(C_ERROR "unknown data!, no handler!" C_OFF "\n");
            bad_syntax();
            exit(1);
    }
    printf(C_ERROR "Error, should not got here!\n report to bugzilla." C_OFF
           "\n");
    exit(22);
}

static int send_and_handle(const char c, const char *l, const char *o)
{
    result_desc *result = NULL;

    /*printf("send_and_handle(%c, %s, %s).\n", c, l, o); */
    if (!(result = send_command(c, l, o)))
    {
        /*print_out("Command -%c --%s, %s - Unable to send command!\n", c, l, o); */
        return (FALSE);
    }
    if (!handle_result(result))
    {
        /*print_out("Command -%c --%s, %s - Bad result!\n", c, l, o); */
        return (FALSE);
    }

    if (result)
        free(result);
    /* if it is an -o option */
    return (TRUE);
}

/* THIS IS MAIN */
int main(int argc, char *argv[])
{
    int cc = 1;

    assert(argv[0]);

    header_printed = FALSE;
#ifdef HAVE_NCURSES
    /* get cols and lines */
    setupterm(NULL, fileno(stdout), (int *) 0);
    nrows = tigetnum((char *) "lines");     /* to work around a compiler warning */
    ncolumns = tigetnum((char *) "cols");   /* (iffy function definitions) */
    printf("This terminal has %i cols and %i rows\n", ncolumns, nrows);
#endif
    if (argv[1] && strcmp(argv[1], "--quiet") == 0)
        quiet = TRUE;

    /* print banner */
    print_out(C_FG_LIGHT_BLUE
              "\n Next Generation init Control. version ( %s )" C_OFF "\n",
              VERSION);
    print_out(C_FG_NEON_GREEN " http://initng.thinktux.net" C_OFF "\n");
    print_out(" Author: Jimmy Wennlund <jimmy.wennlund@gmail.com>\n\n");


    /*
     * Skip path in Argv.
     * example argv[0] == "/sbin/ngc" then Argv == "ngc"
     * example argv[0] == "./ngstart" then Argv == "ngstart"
     */
    {
        int i = 0;

        while (argv[0][i])
            i++;

        while (i > 0 && argv[0][i] != '/')
        {
            i--;
        }

        Argv = &argv[0][i];

        while (Argv[0] == '/')
            Argv++;

        /*printf("Argv: %s\n", Argv); */
    }

    /* check if cmd-line contains "ngdc", if set debug */
    if (strncasecmp(Argv, "ngd", 3) == 0)
    {
        print_out(C_FG_YELLOW "Warning. This is ngdc!" C_OFF "\n");
        debug = TRUE;
    }


    if (debug == FALSE && getuid() != 0)
    {
        printf(C_ERROR "You need root access to communicate with initng."
               C_OFF "\n");
        exit(2);
    }

    /* check that Argc string is not ngc, NGC, ngdc, NGDC */
    if (strcasecmp(Argv, "ngc") != 0 && strcasecmp(Argv, "ngdc") != 0)
    {
        /*printf("Not standard ngc or ngdc command.\n"); */
        /* check if argv0 starts with "ng..." */
        if (Argv[0] == 'n' && Argv[1] == 'g' && Argv[2])
        {
            /* if its ngstart, call send_and_handle("start"); */
            if (Argv[2] == 'd' && Argv[3])
            {
                /* check if is a string or char */
                if (Argv[4])
                    return (send_and_handle('\0', &Argv[3], argv[1]));
                else
                    return (send_and_handle(Argv[3], NULL, argv[1]));
            }
            else
            {
                /* check if its a string or char */
                if (Argv[3])
                    return (send_and_handle('\0', &Argv[2], argv[1]));
                else
                    return (send_and_handle(Argv[2], NULL, argv[1]));
            }
        }

        /* else call standard */
        return (send_and_handle('\0', Argv, argv[1]));
    }

    /* make sure there are any arguments at all */
    if (argc <= 1)
    {
        print_usage();
        exit(0);
    }

    while (argv[cc])
    {
        char *opt = NULL;

        /* every fresh start needs a '-' char */
        if (argv[cc][0] != '-')
        {
            bad_syntax();
            exit(1);
        }

        /* check that there is a char after the '-' */
        if (!argv[cc][1])
        {
            bad_syntax();
            exit(1);
        }

        /* if next option contains data, but not starting with an '-', its considered an option */
        if (argv[cc + 1] && argv[cc + 1][0] != '-')
            opt = argv[cc + 1];


        /* if it is an --option */
        if (argv[cc][1] == '-')
        {
            if (!send_and_handle('\0', &argv[cc][2], opt))
                exit(1);
        }
        else
        {
            if (!send_and_handle(argv[cc][1], NULL, opt))
                exit(1);
        }

        cc++;
        if (opt)
            cc++;


    }
    printf("\n\n");
    exit(0);
}

int ngc2_recvall(int sock, void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = recv(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 0;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 1;
}


/* get help */
int ngc_hlp(void)
{
    /* data got */
    help_row row;

    /* print head */
    print_out(" ngc understand this commands:\n\n");
    print_out(" short Option                          : description\n");
    print_out(" ----------------------------------------------------------\n");

    while (ngc2_recvall(sock, &row, sizeof(help_row)))

    {
        char lname[65];

        /* copy name to the new static array */
        strncpy(lname, row.l, 64);

        switch (row.o)
        {
            case USES_OPT:
                strcat(lname, C_FG_YELLOW " <opt>" C_OFF);
                break;
            case REQUIRES_OPT:
                strcat(lname, C_FG_CYAN " opt" C_OFF);
                break;
            default:
                strcat(lname, C_FG_CYAN C_OFF);
                break;
        }

        /* to prevent an possible overflow */
        lname[64] = '\0';

        if (row.c != '\0')
        {
            print_out(" [" C_FG_LIGHT_RED "-%c" C_OFF "] --%-40s: %s\n",
                      row.c, lname, row.d);
        }
        else
        {
            print_out("       --%-40s: %s\n", lname, row.d);
        }
    }
    return (TRUE);
}


int ngc_active_db(void)
{
    active_row row;
    struct timeval now;

    gettimeofday(&now, NULL);

    /* print head */
#ifdef SIENCE
    print_out("hh:mm:ss service                                :status                   since\n");
    /* don't make it weighter! only 80chars, not weighter. */
    print_out("--------------------------------------------------------------------------------\n");
#else
    print_out(C_FG_LIGHT_RED "hh:mm:ss" C_OFF
              " service                               : status\n");
    /* don't make it weighter! only 80chars, not weighter. */
    print_out("----------------------------------------------------------------\n");
#endif

    while (ngc2_recvall(sock, &row, sizeof(active_row)))
    {
        struct tm *ts = localtime(&row.t.tv_sec);

#ifdef SIENCE
        int tbuf = MS_DIFF(now, row.t) / 1000;
        char tch = 's';

        if (tbuf > 60)
        {
            tbuf /= 60;
            if (tbuf > 60)
            {
                tbuf /= 60;
                if (tbuf > 24)
                {
                    tbuf /= 24;
                    tch = 'd';
                }
                else
                    tch = 'h';
            }
            else
                tch = 'm';
        }
        /* don't make it weighter! only 80chars, not weighter. */
        print_out("%.2i:%.2i:%.2i %-37s : %-22s %7i%c\n",
                  ts->tm_hour, ts->tm_min, ts->tm_sec, row.n, row.s, tbuf,
                  tch);
#else
        /* don't make it weighter! only 80chars, not weighter. */
        print_out(C_FG_LIGHT_RED "%.2i:%.2i:%.2i" C_OFF " %-37s : ",
                  ts->tm_hour, ts->tm_min, ts->tm_sec, row.n);

        switch (row.i)
        {
            case IS_UP:
                print_out(C_FG_NEON_GREEN "%s" C_OFF "\n", row.s);
                break;
            case IS_DOWN:
                print_out(C_FG_LIGHT_BLUE "%s" C_OFF "\n", row.s);
                break;
            case IS_FAILED:
                print_out(C_FG_LIGHT_RED "%s" C_OFF "\n", row.s);
                break;
            case IS_STARTING:
                print_out(C_FG_YELLOW "%s" C_OFF "\n", row.s);
                break;
            case IS_STOPPING:
                print_out(C_FG_CYAN "%s" C_OFF "\n", row.s);
                break;
            case IS_WAITING:
                print_out(C_FG_MAGENTA "%s" C_OFF "\n", row.s);
                break;
            case IS_UNKNOWN:
                print_out("%s\n", row.s);
                break;
        }
#endif
    }
    close_socket();
    return (TRUE);
}


int ngc_option_db(void)
{
    option_row row;
    char ct[20];

    /* print head */
    print_out(" " C_FG_LIGHT_RED "%-10s" C_OFF C_FG_CYAN "%-8s" C_OFF
              " %-24s %s\n", "Where", "Type", "Name", "Description");
    print_out(" ----------------------------------------------------------------\n");

    while (ngc2_recvall(sock, &row, sizeof(option_row)))
    {
        switch (row.t)
        {
            case STRING:
                strcpy(ct, "STRING");
                break;
            case VARIABLE_STRING:
                strcpy(ct, "V_STRING");
                break;
            case STRINGS:
                strcpy(ct, "STRINGS");
                break;
            case VARIABLE_STRINGS:
                strcpy(ct, "V_STRINGS");
                break;
            case SET:
                strcpy(ct, "SET");
                break;
            case VARIABLE_SET:
                strcpy(ct, "V_SET");
                break;
            case INT:
                strcpy(ct, "INT");
                break;
            case VARIABLE_INT:
                strcpy(ct, "V_INT");
                break;
            case ALIAS:
                strcpy(ct, "ALIAS");
                break;
            case U_D_T:
                strcpy(ct, "U_D_T");
                break;
            case TIME_T:
                strcpy(ct, "TIME_T");
                break;
            case VARIABLE_TIME_T:
                strcpy(ct, "V_TIME_T");
                break;
        }

        print_out(" " C_FG_LIGHT_RED "%-10s" C_OFF C_FG_CYAN "%-8s" C_OFF
                  " %-24s %s\n", row.o, ct, row.n, row.d);

    }
    print_out("\n");
    close_socket();
    return (TRUE);
}


/* Avoid using goto */
void bad_syntax(void)
{
    printf(C_ERROR "Bad syntax!\n" C_OFF "\n");
    print_usage();
}
