/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005-6 Aidan Thornton <makomk@lycos.co.uk>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include "ngcs_common.h"
#include <sys/socket.h>
#include <sys/un.h>
#include <assert.h>
#include <unistd.h>

int ngcs_sendmsg(int sock, int chan, int type, int len, const char *data)
{
    int head[3];

    if (len > 0)
        assert(data);
    head[0] = chan;
    head[1] = type;
    head[2] = len;
    if (ngcs_sendall(sock, head, 3 * sizeof(int)))
        return 1;
    if (len <= 0)
        return 0;
    if (ngcs_sendall(sock, data, len))
        return 1;
    return 0;
}

int ngcs_recvmsg(int sock, int *chan, int *type, int *len, char **data)
{
    int head[3];

    assert(chan);
    assert(type);
    assert(len);
    assert(data);

    if (ngcs_recvall(sock, head, 3 * sizeof(int)))
        return 1;
    *chan = head[0];
    *type = head[1];
    *len = head[2];
    if (*len < 0)
    {
        *data = NULL;
        return 0;
    }
    *data = malloc(*len);
    if (*data == NULL)
        return 1;
    if (ngcs_recvall(sock, *data, *len))
        return 1;
    return 0;
}

int ngcs_pack(ngcs_data * data, int cnt, char *buf)
{
    int n;
    int outcnt = 0;

    for (n = 0; n < cnt; n++)
    {
        if (buf)
        {
            *(int *) buf = data[n].type;
            buf += sizeof(int);
        }
        switch (data[n].type)
        {
            case NGCS_TYPE_INT:
                outcnt += 3 * sizeof(int);
                if (buf)
                {
                    *(int *) buf = sizeof(int);
                    buf += sizeof(int);
                    *(int *) buf = data[n].d.i;
                    buf += sizeof(int);
                }
                break;
            case NGCS_TYPE_LONG:
                outcnt += 2 * sizeof(int) + sizeof(long);
                if (buf)
                {
                    *(int *) buf = sizeof(long);
                    buf += sizeof(int);
                    *(int *) buf = data[n].d.l;
                    buf += sizeof(long);
                }
                break;
            case NGCS_TYPE_STRING:
                if (data[n].len < 0)
                    data[n].len = strlen(data[n].d.s);
                outcnt += 2 * sizeof(int) + data[n].len;
                if (buf)
                {
                    *(int *) buf = data[n].len;
                    buf += sizeof(int);
                    if (data[n].len)
                        memcpy(buf, data[n].d.s, data[n].len);
                    buf += data[n].len;
                }
                break;
            default:
                if (data[n].len < 0)
                    return 1;
                outcnt += 2 * sizeof(int) + data[n].len;
                if (buf)
                {
                    *(int *) buf = data[n].len;
                    buf += sizeof(int);
                    if (data[n].len)
                        memcpy(buf, data[n].d.p, data[n].len);
                    buf += data[n].len;
                }
                break;
        }
    }
    return outcnt;
}


int ngcs_unpack(const char *data, int len, ngcs_data ** res)
{
    const char *d = data;
    int l = len;
    int cnt = 0;
    while (l >= 2 * (int) sizeof(int))
    {
        int head[2];
        memcpy(head, d, 2 * sizeof(int));
        if (head[1] < 0)
            return -1;
        d += 2 * sizeof(int) + head[1];
        l -= 2 * sizeof(int) + head[1];
        cnt++;
    }
    d = data;
    l = len;
    *res = malloc(cnt * sizeof(ngcs_data));
    cnt = 0;
    while (l >= 2 * (int) sizeof(int))
    {
        int head[2];
        memcpy(head, d, 2 * sizeof(int));
        (*res)[cnt].type = head[0];
        (*res)[cnt].len = head[1];
        if (head[1] < 0)
        {
            free(*res);
            *res = NULL;
            return -1;
        }
        d += 2 * sizeof(int);
        l -= 2 * sizeof(int);
        if (l < (*res)[cnt].len)
        {
            free(*res);
            *res = NULL;
            return -1;
        }
        switch ((*res)[cnt].type)
        {
            case NGCS_TYPE_INT:
                if ((*res)[cnt].len != sizeof(int))
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                (*res)[cnt].d.i = *(int *) d;
                break;
            case NGCS_TYPE_LONG:
                if ((*res)[cnt].len != sizeof(long))
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                (*res)[cnt].d.i = *(int *) d;
                break;
            case NGCS_TYPE_STRING:
                (*res)[cnt].d.s = malloc((*res)[cnt].len + 1);
                if (!(*res)[cnt].d.s)
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                memcpy((*res)[cnt].d.s, d, (*res)[cnt].len);
                (*res)[cnt].d.s[(*res)[cnt].len] = 0;
                break;
            default:
                (*res)[cnt].d.p = malloc((*res)[cnt].len);
                if (!(*res)[cnt].d.p)
                {
                    free(*res);
                    *res = NULL;
                    return -1;
                }
                memcpy((*res)[cnt].d.p, d, (*res)[cnt].len);
                break;
        }
        d += (*res)[cnt].len;
        l -= (*res)[cnt].len;
        cnt++;
    }
    return cnt;
}

void ngcs_free_unpack(int len, ngcs_data * res)
{
    int n;

    for (n = 0; n < len; n++)
    {
        switch (res[n].type)
        {
            case NGCS_TYPE_INT:
            case NGCS_TYPE_LONG:
                break;
            case NGCS_TYPE_STRING:
                if (res[n].d.s)
                    free(res[n].d.s);
                break;
            default:
                if (res[n].d.p)
                    free(res[n].d.p);
                break;
        }
    }
    free(res);
}

ngcs_conn* ngcs_conn_from_fd(int fd, void* userdata, 
                             void (*close_hook)(ngcs_conn* conn)) 
{
    ngcs_conn* conn;
    struct timeval tv;

    conn = malloc(sizeof(ngcs_conn));
    if(conn == NULL)
        return NULL;

    /* FIXME - should set non-blocking and use buffering */
    tv.tv_sec = 2;
    tv.tv_usec = 0;
    if (setsockopt(fd, SOL_SOCKET, SO_RCVTIMEO, &tv, sizeof(tv)) ||
        setsockopt(fd, SOL_SOCKET, SO_SNDTIMEO, &tv, sizeof(tv)))
    {
        /* FIXME - shoupd print error? */
        return NULL;
    }
    

    INIT_LIST_HEAD(&conn->chans.list);
    conn->fd = fd;
    conn->userdata = userdata;
    conn->close_hook = close_hook;
    return conn;
}


ngcs_chan *ngcs_chan_reg(ngcs_conn * conn, int chanid,
                               void (*gotdata) (ngcs_chan *, int, int,
                                                 char *),
                                void (*chanfree) (ngcs_chan *))
{
    ngcs_chan *chan = malloc(sizeof(ngcs_chan));

    if (!chan)
        return NULL;

    chan->id = chanid;
    chan->conn = conn;
    chan->user_data = NULL;
    chan->gotdata = gotdata;
    chan->free = chanfree;
    chan->list.next = 0;
    chan->list.prev = 0;
    list_add(&chan->list, &conn->chans.list);
    return chan;
}

void ngcs_chan_del(ngcs_chan * chan)
{
    if (chan->conn->fd >= 0 && chan->free)
        chan->free(chan);
    list_del(&chan->list);
    free(chan);
}


void ngcs_conn_data_ready(ngcs_conn* conn)
{
    int chanid, len, type;
    char *data;
    ngcs_chan *chan;
    
    /* Get message */
    if (ngcs_recvmsg(conn->fd, &chanid, &type, &len, &data))
    {
        ngcs_conn_close(conn);
        return;
    }

    /* Dispatch to handler for this channel */
    while_ngcs_chans(chan, conn) if (chan->id == chanid)
    {
        if (chan->gotdata)
            chan->gotdata(chan, type, len, data);
        free(data);
        return;
    }

    /* Unknown channel - close it */
    if (ngcs_sendmsg(conn->fd, chanid, NGCS_TYPE_NULL, -1, NULL))
        ngcs_conn_close(conn);
    free(data);
    return;
}

void ngcs_conn_close(ngcs_conn* conn)
{
    ngcs_chan *chan, *tmp;
    if(conn->fd < 0) return;
    close(conn->fd);
    conn->fd = -1;
    while_ngcs_chans_safe(chan, tmp, conn)
    {
        if (chan->free)
            chan->free(chan);
    }
    if(conn->close_hook)
        conn->close_hook(conn);
}

void ngcs_conn_free(ngcs_conn* conn)
{
    ngcs_chan *chan, *tmp;
    ngcs_conn_close(conn);
    while_ngcs_chans_safe(chan, tmp, conn)
    {
        list_del(&chan->list); free(chan);
    }
    free(conn);
    /* TODO */
}

int ngcs_sendall(int sock, const void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = send(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}

int ngcs_recvall(int sock, void *buf, int len)
{
    int ret;

    while (len > 0)
    {
        ret = recv(sock, buf, len, 0);
        if (ret <= 0)
        {
            return 1;
        }
        else
        {
            buf += ret;
            len -= ret;
        }
    }
    return 0;
}

int ngcs_chan_send(ngcs_chan * chan, int type, int len, const char *data)
{
    if (chan->conn->fd < 0)
        return 1;
    if (ngcs_sendmsg(chan->conn->fd, chan->id, type, len, data))
    {
        ngcs_conn_close(chan->conn);
        return 1;
    }
    return 0;
}
