
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <initng.h>
#define _GNU_SOURCE
#include <stdio.h>
#include <alloca.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>
#include <stdlib.h>
#include <assert.h>

#include <initng_global.h>
#include <initng_service_cache.h>
#include <initng_string_tools.h>
#include <initng_open_read_close.h>
#include <initng_toolbox.h>
#include <initng_plugin_hook.h>
#include <initng_handler.h>
#include <initng_static_service_types.h>
#include <initng_service_types.h>
#include <initng-paths.h>

stype_h *TYPE_RUNLEVEL;

/* a simple parser for a runlevel file */
static service_cache_h *initng_rl_parser(const char *runlevel_name)
{
    service_cache_h *n_service; /* service struct pointer too   */
    char *filetoparse = NULL;
    char *w_depends = NULL;
    char *w = NULL;
    char *a = NULL;

    assert(runlevel_name);

    /* make sure service type RUNLEVEL is set */
    if (!TYPE_RUNLEVEL)
    {
        TYPE_RUNLEVEL = initng_service_types_get("runlevel");
        if (!TYPE_RUNLEVEL)
        {
            F_("ERROR, runlevel servicetype is not found, make sure runlevel plugin is loaded.\n");
            return (NULL);
        }
    }
    filetoparse = (char *) i_calloc(strlen(INITNG_ROOT) + 1 +
                                    strlen(runlevel_name) + 10, sizeof(char));

    D_("initng_rl_parser(%s);\n", runlevel_name);
    strcpy(filetoparse, INITNG_ROOT "/");
    strcat(filetoparse, runlevel_name);
    strcat(filetoparse, ".runlevel");
    D_("initng_rl_parser(%s): loading file %s.\n",
       runlevel_name, filetoparse);

    /* allocate a new service */
    if (!(n_service = initng_service_cache_new(runlevel_name, TYPE_RUNLEVEL)))
    {
        free(filetoparse);
        return (NULL);
    }

    /* copy file filename into allocated file_buf */
    if (!open_read_close(filetoparse, &w_depends))
    {
        D_("parse_file(%s): Can't open config file!\n", filetoparse);
        if (n_service->name)
            free(n_service->name);

        /* free all data entries */
        initng_service_cache_remove_all(n_service);

        if (n_service)
            free(n_service);
        free(filetoparse);
        return (NULL);
    }
    w = w_depends;

    while (w[0] != '\0')
    {
        /* skip leading spaces */
        JUMP_SPACES(w);
        if (w[0] == '\0')
            break;

        /* skip lines, starting with '#' */
        if (w[0] == '#')
        {
            JUMP_TO_NEXT_ROW(w);
            continue;
        }

        a = st_dup_line(&w);
        if (a)
        {
            D_("adding dep: \"%s\"\n", a);
            inting_service_cache_set_another_string(&NEED, n_service, a);
        }

        JUMP_TO_NEXT_ROW(w);
    }

    free(w_depends);

    initng_service_cache_set_string(&FROM_FILE, n_service,
                                    i_strdup(filetoparse));

    free(filetoparse);

    if (initng_service_cache_add(n_service))
        return (n_service);
    return (NULL);
}

int module_init(int api_version)
{
    /* initziate globals */
    TYPE_RUNLEVEL = NULL;

    D_("initng_rl_parser: module_init();\n");
    if (api_version != API_VERSION)
    {
        F_("This module is compiled for api_version %i version and initng is compiled on %i version, won't load this module!\n", INITNG_VERSION, api_version);
        return (FALSE);
    }

    return (initng_plugin_hook_add(&g.PARSERS, 60, &initng_rl_parser));
}

void module_unload(void)
{
    initng_plugin_hook_del(&g.PARSERS, &initng_rl_parser);
}
