
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#ifndef ACTIVE_DB_H
#define ACTIVE_DB_H

typedef struct active_type active_db_h;


#include <sys/types.h>
#include <unistd.h>
#include "initng_service_cache.h"

#include "initng_struct_data.h"
#include "initng_list.h"

#include "initng_service_types.h"
#include "initng_active_state.h"
#include "initng_process_db.h"

#define MAX_SUCCEEDED 30

/* the active service struct */
struct active_type
{
    /* identification */
    char *name;
    stype_h *type;

    /* service data */
    service_cache_h *from_service;

    /* current state */
    a_state_h *current_state;
    struct timeval time_current_state;  /* the time got current state */

    /* last    state */
    a_state_h *last_state;
    struct timeval time_last_state;     /* the time got last state */

    /* Rough  state */
    e_is last_rought_state;
    struct timeval last_rought_time;

    /* processes */
    process_h processes;

    /* list container, for data */
    s_data data;

    /* Alarm, the current state alarm is run when this time passes */
    time_t alarm;

    /* depend cache - Optimizion to speed up UP_DEPS_CHECK */
    int depend_cache;

    /* the list */
    struct list_head list;
};

/* allocate */
active_db_h *initng_active_db_new(const char *name);

/* searching */
active_db_h *initng_active_db_find_by_exact_name(const char *service);
active_db_h *initng_active_db_find_by_name(const char *service);
active_db_h *initng_active_db_find_in_name(const char *service);
active_db_h *initng_active_db_find_by_pid(pid_t pid);
active_db_h *initng_active_db_find_by_service_h(service_cache_h * service);

/* mangling */
void initng_active_db_change_service_h(service_cache_h * from,
                                       service_cache_h * to);
void initng_active_db_compensate_time(time_t skew);

/* the db */
#define initng_active_db_del(serv) list_del(&(serv)->list)
int initng_active_db_add(active_db_h * new_a);
int initng_active_db_count(a_state_h * state);
void initng_active_db_free(active_db_h * pf);
void initng_active_db_free_all(void);

/* dependency check */
/* returns TRUE if service DEP ON check */
/*int initng_active_db_dep_on(active_db_h * service, active_db_h * check);
   int initng_active_db_dep_on_deep(active_db_h * service, active_db_h * check); */

/* utils */
int initng_active_db_percent_started(void);
int initng_active_db_percent_stopped(void);

/* get/set data handling */
#define initng_active_db_get_int(t,a) initng_active_db_get_int_var(t, NULL, a)
#define initng_active_db_get_string(t,a) initng_active_db_get_string_var(t, NULL, a)
#define initng_active_db_get_next_string(t,a,l) initng_active_db_get_next_string_var(t,NULL, a, l)


s_data *initng_active_db_get_next(s_entry * type, active_db_h * from_active,
                                  s_data * last);

#define initng_active_db_is(t,a) initng_active_db_is_var(t, NULL, a)
int initng_active_db_get_int_var(s_entry * type, const char *vn,
                                 active_db_h * from_active);
const char *initng_active_db_get_string_var(s_entry * type, const char *vn,
                                            active_db_h * from_active);
const char *initng_active_db_get_next_string_var(s_entry * type,
                                                 const char *vn,
                                                 active_db_h * from_active,
                                                 const char *last);
int initng_active_db_is_var(s_entry * type, const char *vn,
                            active_db_h * from_active);

/* count the total number of this type */
int initng_active_db_count_type(s_entry * type, active_db_h * from_active);
void initng_active_db_clean_down(void);

#define initng_active_db_set_string(type, act, string)  d_set_string(type, &act->data.list, string)
#define initng_active_db_set_another_string(type, act, string) d_set_another_string(type, &act->data.list, string)
#define initng_active_db_set_int(type, act, value) d_set_int(type, &act->data.list, value)
#define initng_active_db_set(type, act) d_set(type, &act->data.list)
#define initng_active_db_set_string_var(type, vn, act, string)  d_set_string_var(type, vn, &act->data.list, string)
#define initng_active_db_set_another_string_var(type, vn, act, string) d_set_another_string_var(type, vn, &act->data.list, string)
#define initng_active_db_set_int_var(type, vn, act, value) d_set_int_var(type, vn, &act->data.list, value)
#define initng_active_db_set_var(type, vn, act) d_set_var(type, vn, &act->data.list)

#define initng_active_db_remove(type, act) d_remove(type, &act->data.list)
#define initng_active_db_remove_var(type, vn, act) d_remove(type, vn, &act->data.list)
#define initng_active_db_remove_all(act) d_remove_all(&act->data.list)
#define initng_active_db_unset(type, act) initng_active_db_remove(type, act)

#define while_active_db(current) list_for_each_entry_prev(current, &g.active_database.list, list)
#define while_active_db_safe(current, safe) list_for_each_entry_prev_safe(current, safe, &g.active_database.list, list)

#endif
