
/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

/* common standard first define */
#include "initng.h"

/* system headers */
#include <time.h>                           /* time() */
#include <fcntl.h>                          /* fcntl() */
#include <sys/un.h>                         /* memmove() strcmp() */
#include <sys/wait.h>                       /* waitpid() sa */
#include <linux/kd.h>                       /* KDSIGACCEPT */
#include <sys/ioctl.h>                      /* ioctl() */
#include <stdio.h>                          /* printf() */
#include <stdlib.h>                         /* free() exit() */
#include <sys/reboot.h>                     /* reboot() RB_DISABLE_CAD */
#include <assert.h>
#include <ctype.h>                          /* isgraph */

/* own header */
#include "initng_env_variable.h"

/* local headers include */
#include "initng_main.h"
#include "initng_active_db.h"
#include "initng_load_module.h"
#include "initng_plugin_callers.h"
#include "initng_common.h"
#include "initng_toolbox.h"
#include "initng_string_tools.h"
#include "initng_global.h"

#include <initng-paths.h>

const char *initng_environ[] = {
    "INITNG=" INITNG_VERSION,
    "INITNG_CREATOR=" INITNG_CREATOR,
    "INIT_VERSION=" INITNG_VERSION,
    "INITNG_PLUGIN_DIR=" INITNG_PLUGIN_DIR,
    "INITNG_ROOT=" INITNG_ROOT,
    "PATH=/bin:/usr/bin:/sbin:/usr/sbin:/usr/local/bin:/opt/bin",
    "HOME=/root",
    "USER=root",
    "TERM=linux",
    NULL
};

/* creates a custom set of environment variables, for passing to exec
 * Warning: this leaks memory, so only use if you are going to exec()!
 * Don't even *think* about free()ing the strings, unless you want
 * invalid pointers in the service_db! (easy to fix, see comment below)
 */
char **new_environ(active_db_h * s)
{
    int allocate;
    int nr = 0;
    int i;
    char **env;
    char *var;

    /* Better safe than sorry... */
    if (s && s->name == NULL)
        s->name = i_strdup("unknown");

    /*
     * FIRST, try to figure out how big array we want to create.
     */

    /*  At least 11 allocations below, and place for plugin added to (about 100) */
    allocate = 114;

    /* count existing env's */
#ifdef USE_OLD_ENV                          /* DO NOT USE - BROKEN */
    while (environ[nr])
        allocate++;
#endif

    /* count ENVs in service ENV variable */
    if (s)
        allocate += initng_active_db_count_type(&ENV, s);


    /* finally allocate */
    env = (char **) i_calloc(allocate, sizeof(char *));

    /* duplicate */
#ifdef USE_OLD_ENV
    for (nr = 0; environ[nr] && nr < allocate; nr++)
    {
        env[nr] = i_strdup(environ[nr]);
    }
#endif

    /* add all static defined above in initng_environ */
    for (nr = 0; initng_environ[nr]; nr++)
    {
        env[nr] = i_strdup(initng_environ[nr]);
    }


    if (s && (nr + 4) < allocate)
    {
        env[nr] = (char *) i_calloc(1, sizeof(char) * (9 + strlen(s->name)));
        strcpy(env[nr], "SERVICE=");
        strcat(env[nr], s->name);
        nr++;

        env[nr] = (char *) i_calloc(1, sizeof(char) * (6 + strlen(s->name)));
        strcpy(env[nr], "NAME=");
        strcat(env[nr], st_strip_path(s->name));
        nr++;

        env[nr] = (char *) i_calloc(1, sizeof(char) * (7 + strlen(s->name)));
        strcpy(env[nr], "CLASS=");
        strcat(env[nr], s->name);
        st_strip_end(&env[nr]);
        nr++;

        if (g.dev_console)
        {
            env[nr] = (char *) i_calloc(1,
                                        sizeof(char) * (9 +
                                                        strlen(g.
                                                               dev_console)));
            strcpy(env[nr], "CONSOLE=");
            strcat(env[nr], g.dev_console);
            nr++;
        }
        else
        {
            env[nr] = (char *) i_calloc(1,
                                        sizeof(char) * (9 +
                                                        strlen
                                                        (INITNG_CONSOLE)));
            strcpy(env[nr], "CONSOLE=");
            strcat(env[nr], INITNG_CONSOLE);
            nr++;
        }

        if (g.runlevel && (nr + 1) < allocate)
        {
            env[nr] = (char *) i_calloc(1,
                                        sizeof(char) * (10 +
                                                        strlen(g.runlevel)));
            strcpy(env[nr], "RUNLEVEL=");
            strcat(env[nr], g.runlevel);
            nr++;
        }

        if (g.old_runlevel && (nr + 1) < allocate)
        {
            env[nr] = (char *) i_calloc(1,
                                        sizeof(char) * (14 +
                                                        strlen(g.
                                                               old_runlevel)));
            strcpy(env[nr], "PREVLEVEL=");
            strcat(env[nr], g.old_runlevel);
            nr++;
        }

        env[nr] = NULL;
        /* insert all env strings from config */
        {
            var = NULL;
            s_data *tmp = NULL;

            while ((nr + 1) < allocate)
            {
                char *fixed = NULL;

                if (!(tmp = initng_active_db_get_next(&ENV, s, tmp)))
                    break;

                fixed = fix_variables(tmp->t.s, s);

                /* then malloc */
                var = i_calloc((strlen(tmp->vn) + strlen(fixed) + 3),
                               sizeof(char));
                if (!var)
                    continue;

                /* copy the data */
                strcpy(var, tmp->vn);
                strcat(var, "=");
                strcat(var, fixed);

                /* free the fixed ones */
                free(fixed);

                /* check for duplicates */
                for (i = 0; i < nr; i++)
                {
                    if (is_same_env_var(env[i], var))
                    {
                        /* we may want to override environemntal variables
                           set above, particularly PATH and HOME */
                        free(env[i]);
                        env[i] = var;
                        break;
                    }
                }

                if (i == nr)
                    env[nr++] = var;
            }
        }
    }

    /* null last */
    if (env[nr] != NULL)
        env[nr++] = NULL;

#ifdef DEBUG
    for (nr = 0; env[nr]; nr++)
    {
        D_("environ[%i] = \"%s\"\n", nr, env[nr]);
    }
#endif

    /* return new environ */
    return (env);
}

/* this frees an environment variable - not to be used on the output of
 * new_environ!
 */
void free_environ(char **tf)
{
    int i = 0;

    D_("free_environ();\n");
    assert(tf);
    for (i = 0; tf[i]; tf++)
    {
        free(tf[i]);
    }
    free(tf);
}

int is_same_env_var(char *var1, char *var2)
{
    int i = 0;

    if (!var1 || !var2)
        return 0;                           /* bad error checking in caller! */

    for (i = 0; var1[i] && var2[i] && var1[i] != '=' && var1[i] == var2[i];
         i++)
        ;

    return var1[i] == var2[i];
}

/* Handles redefinitions like "env FOO = $FOO bar" */
char *fix_redefined_variable(const char *name, const char *oldval,
                             const char *newdef)
{
    char *to;                   /* this is the pointer we are going to return */
    char *set;                  /* pointer to step */

    assert(name);
    assert(oldval);
    assert(newdef);

    /* make sure we got any data */
    if (!newdef)
        return (NULL);

    /* allocate that much memory that we think we need */
    /* TODO, add checks that this value is never overridden */
    to = (char *) i_calloc((strlen(newdef) + 150), sizeof(char));
    if (!to)
        return (NULL);

    set = to;
    /* while there is data to parse */
    while (newdef[0])
    {
        int len = 0;

        /*
         * if its not containing the magical char loop, just copy the
         * char and continue the main loop.
         */
        if (newdef[0] != '$')
        {
            set[0] = newdef[0];
            newdef++;
            set++;
            set[0] = '\0';
            continue;
        }


        /* jump the '$' char */
        newdef++;
        if (!newdef[0])
            continue;

        /*
         * Calculate length of variable ${VARIABLE}.
         */
        if (newdef[0] == '{')               /* if this is a ${VARIABLE} */
        {
            newdef++;
            for (; newdef[len] && newdef[len] != '}'; len++) ;
        }
        else                                /* else a $VARIABLE */
        {
            for (; isgraph(newdef[len]); len++) ;
        }

        if (strncasecmp(newdef, name, len) == 0 && (int) strlen(name) == len)
        {
            strcpy(set, oldval);
            /* go forward */
            newdef += len;
            if (newdef[0] == '}')
                newdef++;
            while (set[0])
                set++;
        }
        else
        {
            if (newdef[-1] == '{')
            {
                strncpy(set, newdef - 2, len + 2);
                newdef += len;
                set += len + 2;
                set[0] = '\0';
            }
            else
            {
                strncpy(set, newdef - 1, len + 1);
                newdef += len;
                set += len + 1;
                set[0] = '\0';
            }
        }
    }

    /* make sure end is terminated */
    set[0] = '\0';

    /* return the string we now fixed */
    return (to);

}

/* This function will change variables in from, like $NAME, $SERVICE, $CLASS
 * into its right content, and return a new allocated string with the result.
 */
char *fix_variables(const char *from, active_db_h * s)
{
    char *to;                   /* this is the pointer we are going to return */
    char *set;                  /* pointer to step */
    int i;                      /* temporary counter */

    assert(s);
    assert(s->name);

    /* make sure we got any data */
    if (!from)
        return (NULL);

    /* allocate that much memory that we think we need */
    /* TODO, add checks that this value is never overridden */
    to = (char *) i_calloc((strlen(from) + 150), sizeof(char));
    if (!to)
        return (NULL);

    set = to;
    /* while there is data to parse */
    while (from[0])
    {
        int len = 0;

        /*
         * if its not containing the magical char loop, just copy the
         * char and continue the main loop.
         */
        if (from[0] != '$')
        {
            set[0] = from[0];
            from++;
            set++;
            set[0] = '\0';
            continue;
        }


        /* jump the '$' char */
        from++;
        if (!from[0])
            continue;

        /*
         * Calculate length of variable ${VARIABLE}.
         */
        if (from[0] == '{')                 /* if this is a ${VARIABLE} */
        {
            from++;
            for (; from[len] && from[len] != '}'; len++) ;
        }
        else                                /* else a $VARIABLE */
        {
            for (; isgraph(from[len]); len++) ;
        }

        /*
         * Start comparing keywords.
         */

        if (strncasecmp(from, "NAME", len) == 0)
        {
            const char *tmp = NULL;

            if ((tmp = st_strip_path(s->name)))
            {
                strcpy(set, tmp);
                goto go_forward;
            }
        }

        if (strncasecmp(from, "SERVICE", len) == 0)
        {
            strcpy(set, s->name);
            goto go_forward;
        }

        if (strncasecmp(from, "CLASS", len) == 0)
        {
            char *tmp = NULL;

            if ((tmp = st_get_path(s->name)))
            {
                strcpy(set, tmp);
                free(tmp);
            }
            goto go_forward;
        }

        if (g.runlevel && strncasecmp(from, "CONSOLE", len) == 0)
        {
            if (g.dev_console)
                strcpy(set, g.dev_console);
            else
                strcpy(set, INITNG_CONSOLE);

            goto go_forward;
        }

        if (g.runlevel && strncasecmp(from, "RUNLEVEL", len) == 0)
        {
            strcpy(set, g.runlevel);
            goto go_forward;
        }

        if (g.runlevel && strncasecmp(from, "PREVLEVEL", len) == 0)
        {
            strcpy(set, g.old_runlevel);
            goto go_forward;
        }

        /* Fetch variable from initng_environ */
        for (i = 0; initng_environ[i]; i++)
        {
            if (strncasecmp(from, initng_environ[i], len) == 0
                && initng_environ[i][len] == '=')
            {
                strcpy(set, &initng_environ[i][len + 1]);
                goto go_forward;
            }
        }

        {
            s_data *tmp = NULL;

            while ((tmp = initng_active_db_get_next(&ENV, s, tmp)))
            {
                if (strncasecmp(from, tmp->vn, len) == 0 && tmp->vn[len] == 0)
                {
                    strcpy(set, tmp->t.s);
                    goto go_forward;
                }
            }
        }

        /* Get the variable from the data set on an initng_variable */
        {
            s_entry *entry = NULL;

            /* Copy the string, so that we can put an '\0' on the end */
            char *tmp = i_strndup(from, len);

            if (!tmp)
                goto print_error;

            /* search if a initng variable exits with that name */
            entry = initng_service_data_types_find(tmp);

            /* free the duped one */
            free(tmp);

            /* if we got an entry, and if the entry exits in this service */
            if (entry && initng_active_db_is(entry, s))
            {
                /* Handle differently with different types */
                switch (entry->opt_type)
                {
                        /* if it is a string, just strcpy in it */
                    case STRING:
                        {
                            strcpy(set,
                                   initng_active_db_get_string(entry, s));
                            goto go_forward;
                        }
                    case STRINGS:
                        {
                            const char *tmp2 = NULL;

                            while ((tmp2 =
                                    initng_active_db_get_next_string(entry, s,
                                                                     tmp2)))
                            {
                                strcat(set, tmp2);
                                strcat(set, " ");
                            }
                            if (tmp2)
                                goto go_forward;
                        }
                        /* if it is an integer, sprintf in it */
                    case INT:
                        {
                            sprintf(set, "%i",
                                    initng_active_db_get_int(entry, s));
                            goto go_forward;
                        }
                    default:
                        {
                            D_("Don't know how to set a variable with %s:%i\n", entry->opt_name, entry->opt_type);
                            break;
                        }
                }
            }
        }


      print_error:
        /* Print a fancy error message */
        {
            char *tmp = strndup(from, len);

            W_("Service %s Unset variable ${%s}!\n", s->name, tmp);
            free(tmp);
        }
        /* this will set the value to nothing */
        set[0] = '\0';

      go_forward:
        /* go forward */
        from += len;
        if (from[0] == '}')
            from++;
        while (set[0])
            set++;

    }

    /* make sure end is terminated */
    set[0] = '\0';

    /* return the string we now fixed */
    return (to);
}
