/* Initng, a next generation sysvinit replacement.
 * Copyright (C) 2005 Jimmy Wennlund <jimmy.wennlund@gmail.com>
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General
 * Public License along with this library; if not, write to the
 * Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA  02110-1301, USA.
 */

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include <dlfcn.h>
#include <assert.h>
#include <errno.h>
#include <string.h>
#include <fcntl.h>                          /* fcntl() */
#include <time.h>
#include "initng.h"
#include "initng_global.h"
#include "initng_service_cache.h"
#include "initng_load_module.h"
#include "initng_toolbox.h"
#include "initng_signal.h"

#include "initng_plugin_callers.h"
#include "initng_plugin.h"


/* not editing anything, so it doesn't have to use a recursive function */
void initng_plugin_callers_alarm(void)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.ALARM, q)
    {
        (*current->c.alarm) ();
    }
}

void initng_plugin_callers_hup(void)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.HUP, q)
    {
        (*current->c.hup) ();
    }
}

void initng_plugin_callers_pipe(void)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.PIPE, q)
    {
        (*current->c.hup) ();
    }
}

int initng_plugin_callers_handle_killed(active_db_h * s, process_h * p)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.HANDLE_KILLED, q)
    {
        D_("Calling killed_handle plugin from %s\n", current->from_file);
        if (current->c.handle_killed)
            if (((*current->c.handle_killed) (s, p)) == TRUE)
                return (TRUE);

    }
    return (FALSE);
}


void initng_plugin_callers_compensate_time(int t)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.COMPENSATE_TIME, q)
    {
        if (current->c.swatcher)
        {
            D_("Calling system_state_changed plugin from %s\n",
               current->from_file);
            (*current->c.swatcher) (t);
        }
    }

}

/* called when system state has changed. */
void initng_plugin_callers_load_module_system_changed(h_sys_state state)
{
    s_call *current, *q = NULL;

    while_list_safe(current, &g.SWATCHERS, q)
    {
        if (current->c.swatcher)
        {
            D_("Calling system_state_changed plugin from %s\n",
               current->from_file);
            (*current->c.swatcher) (state);
        }
    }
    return;
}

/* called to dump active_db */
int initng_plugin_callers_dump_state(void)
{
    s_call *current, *q = NULL;
    int retval = FALSE;

    while_list_safe(current, &g.DUMP_STATE, q)
    {
        if (current->c.dump_state)
        {
            D_("Calling dump_state plugin from %s\n", current->from_file);
            retval = (*current->c.dump_state) ();
            if (retval == TRUE)
            {
                D_("dump_state plugin from %s succeeded\n",
                   current->from_file);
                break;
            }
            else if (retval == FAIL)
            {
                F_("dump_state plugin from %s failed\n", current->from_file);
                break;
            }
        }
    }
    return retval;

}

/* called to reload dump of active_db */
int initng_plugin_callers_reload_state(void)
{
    s_call *current, *q = NULL;
    int retval = FALSE;

    while_list_safe(current, &g.RELOAD_STATE, q)
    {
        if (current->c.reload_state)
        {
            D_("Calling reload_state plugin from %s\n", current->from_file);
            retval = (*current->c.reload_state) ();
            if (retval == TRUE)
            {
                D_("reload_state plugin from %s succeeded\n",
                   current->from_file);
                break;
            }
            else if (retval == FAIL)
            {
                F_("reload_state plugin from %s failed\n",
                   current->from_file);
                break;
            }
        }
    }
    return retval;
}
